# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import select
import signal
import socket
import sys
import threading
from types import FrameType
from typing import Any
from wsgiref.simple_server import make_server
from wsgiref.simple_server import WSGIRequestHandler

from oslo_config import cfg
from oslo_log import log as logging
from prometheus_client import make_wsgi_app

from oslo_metrics import message_router


oslo_metrics_configs = [
    cfg.StrOpt(
        'metrics_socket_file',
        default='/var/tmp/metrics_collector.sock',  # noqa: S108
        help='Unix domain socket file to be used to send rpc related metrics',
    ),
    cfg.StrOpt(
        'prometheus_host',
        default='',
        help='Hostname or IP address to serve metrics. An empty value '
        '(the default) makes the server bind to all network '
        'interfaces, equivalent to 0.0.0.0.',
    ),
    cfg.PortOpt(
        'prometheus_port',
        default=3000,
        help='Port number to expose metrics in prometheus format.',
    ),
    cfg.IntOpt(
        'metrics_socket_perm',
        default=0o660,
        help='Permission set to the unix domain socket file',
    ),
    cfg.BoolOpt(
        'wsgi_silent_server',
        default=True,
        help='Whether to silence the WSGI server. If disabled, the '
        'WSGI server will print all requests it receives on '
        'STDOUT. This could be very verbose.',
    ),
]
cfg.CONF.register_opts(oslo_metrics_configs, group='oslo_metrics')


LOG = logging.getLogger(__name__)
CONF = cfg.CONF
logging.register_options(CONF)
logging.setup(CONF, 'oslo-metrics')


class MetricsListener:
    def __init__(self, socket_path: str) -> None:
        self.socket_path = socket_path
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
        self.unlink(socket_path)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind(self.socket_path)
        self.start = True
        self.router = message_router.MessageRouter()

    def unlink(self, socket_path: str) -> None:
        try:
            os.unlink(socket_path)
        except OSError:
            if os.path.exists(socket_path):
                raise

    def serve(self) -> None:
        while self.start:
            readable, writable, exceptional = select.select(
                [self.socket], [], [], 1
            )
            if len(readable) == 0:
                continue
            try:
                LOG.debug("wait for socket.recv")
                # 1 message size should be smaller than 65565
                msg = self.socket.recv(65565)
                LOG.debug("got message")
                self.router.process(msg)
            except TimeoutError:
                pass

    def stop(self) -> None:
        self.socket.close()
        self.start = False


class _SilentHandler(WSGIRequestHandler):
    """WSGI handler that does not log requests."""

    def log_message(self, format: str, *args: Any) -> None:
        """Log nothing."""


httpd = None


def handle_sigterm(_signum: int, _frame: FrameType | None) -> None:
    if httpd is None:
        # this should never happen
        raise RuntimeError('httpd is uninitialized')

    LOG.debug("Caught sigterm")
    shutdown_thread = threading.Thread(target=httpd.shutdown)
    shutdown_thread.start()


def main() -> None:
    cfg.CONF(sys.argv[1:])
    socket_path = cfg.CONF.oslo_metrics.metrics_socket_file
    m = MetricsListener(socket_path)
    try:
        os.chmod(socket_path, cfg.CONF.oslo_metrics.metrics_socket_perm)
    except OSError:
        LOG.error("Changing the mode of the file failed.... continuing")
    mt = threading.Thread(target=m.serve)
    LOG.info("Start oslo.metrics")
    mt.start()

    app = make_wsgi_app()

    global httpd
    if cfg.CONF.oslo_metrics.wsgi_silent_server:
        httpd = make_server(
            CONF.oslo_metrics.prometheus_host,
            CONF.oslo_metrics.prometheus_port,
            app,
            handler_class=_SilentHandler,
        )
    else:
        httpd = make_server(
            CONF.oslo_metrics.prometheus_host,
            CONF.oslo_metrics.prometheus_port,
            app,
        )

    signal.signal(signal.SIGTERM, handle_sigterm)

    try:
        httpd.serve_forever()
    except KeyboardInterrupt:
        pass
    finally:
        LOG.info("Try to stop...")
        os.remove(cfg.CONF.oslo_metrics.metrics_socket_file)
        m.stop()
        httpd.server_close()


if __name__ == "__main__":
    main()



# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import prometheus_client

rpc_server_common_labels = [
    'exchange',
    'topic',
    'server',
    'endpoint',
    'namespace',
    'version',
    'method',
    'process',
]
rpc_client_common_labels = [
    'call_type',
    'exchange',
    'topic',
    'namespace',
    'version',
    'server',
    'fanout',
    'process',
    'method',
    'timeout',
]

rpc_processing_seconds_buckets = [
    0.01,
    0.025,
    0.05,
    0.075,
    0.1,
    0.25,
    0.5,
    0.75,
    1.0,
    2.5,
    5.0,
    7.5,
    10.0,
    25.0,
    50.0,
    75.0,
    100,
]

# RPC Server Metrics
rpc_server_invocation_start_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_invocation_start_total',
    'Total number of RPC invocation start. This doesn\'t count'
    'if rpc server failed to find method from endpoints.',
    rpc_server_common_labels,
)

rpc_server_invocation_end_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_invocation_end_total',
    'Total number of RPC invocation end.',
    rpc_server_common_labels,
)

rpc_server_processing_seconds = prometheus_client.Histogram(
    'oslo_messaging_rpc_server_processing_seconds',
    'Duration of RPC processing.',
    rpc_server_common_labels,
    buckets=rpc_processing_seconds_buckets,
)

rpc_server_exception_total = prometheus_client.Counter(
    'oslo_messaging_rpc_server_exception_total',
    'Total number of exception while RPC processing.',
    rpc_server_common_labels + ['exception'],
)

# RPC Client Metrics
rpc_client_invocation_start_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_invocation_start_total',
    'Total number of RPC invocation start.',
    rpc_client_common_labels,
)

rpc_client_invocation_end_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_invocation_end_total',
    'Total number of RPC invocation end.',
    rpc_client_common_labels,
)

rpc_client_processing_seconds = prometheus_client.Histogram(
    'oslo_messaging_rpc_client_processing_seconds',
    'Duration of RPC processing.',
    rpc_client_common_labels,
    buckets=rpc_processing_seconds_buckets,
)

rpc_client_exception_total = prometheus_client.Counter(
    'oslo_messaging_rpc_client_exception_total',
    'Total number of exception while RPC processing.',
    rpc_client_common_labels + ['exception'],
)



# Copyright 2020 LINE Corp.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
from typing import Any, TypedDict


class UnSupportedMetricActionError(Exception):
    def __init__(self, message: str | None = None) -> None:
        self.message = message


class MetricValidationError(Exception):
    def __init__(self, message: str | None = None) -> None:
        self.message = message


class _MetricActionDict(TypedDict, total=False):
    action: str
    value: str


class MetricAction:
    actions = ['inc', 'observe']

    def __init__(self, action: str, value: str) -> None:
        if action not in self.actions:
            raise UnSupportedMetricActionError(
                f"{action} action is not supported"
            )
        self.action = action
        self.value = value

    @classmethod
    def validate(cls, metric_action_dict: _MetricActionDict) -> None:
        if "value" not in metric_action_dict:
            raise MetricValidationError("action need 'value' field")
        if "action" not in metric_action_dict:
            raise MetricValidationError("action need 'action' field")
        if metric_action_dict["action"] not in cls.actions:
            raise MetricValidationError(
                f"action should be choosen from {cls.actions}"
            )

    @classmethod
    def from_dict(
        cls, metric_action_dict: _MetricActionDict
    ) -> 'MetricAction':
        return cls(metric_action_dict["action"], metric_action_dict["value"])


class Metric:
    def __init__(
        self, module: str, name: str, action: MetricAction, **labels: str
    ) -> None:
        self.module = module
        self.name = name
        self.action = action
        self.labels = labels

    def to_json(self) -> str:
        raw = {
            "module": self.module,
            "name": self.name,
            "action": {
                "value": self.action.value,
                "action": self.action.action,
            },
            "labels": self.labels,
        }
        return json.dumps(raw)

    @classmethod
    def from_json(cls, encoded: str) -> 'Metric':
        metric_dict = json.loads(encoded)
        cls._validate(metric_dict)
        return Metric(
            metric_dict["module"],
            metric_dict["name"],
            MetricAction.from_dict(metric_dict["action"]),
            **metric_dict["labels"],
        )

    @classmethod
    def _validate(cls, metric_dict: dict[str, Any]) -> None:
        if "module" not in metric_dict:
            raise MetricValidationError("module should be specified")

        if "name" not in metric_dict:
            raise MetricValidationError("name should be specified")

        if "action" not in metric_dict:
            raise MetricValidationError("action should be specified")

        if "labels" not in metric_dict:
            raise MetricValidationError("labels should be specified")

        MetricAction.validate(metric_dict["action"])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

sys.path.insert(0, os.path.abspath('../..'))
# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinxcontrib.apidoc',
    'sphinx.ext.intersphinx',
    'openstackdocstheme',
    'oslo_config.sphinxext',
]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/oslo.metrics'
openstackdocs_bug_project = 'oslo.metrics'
openstackdocs_bug_tag = ''

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../oslo_metrics'
apidoc_output_dir = 'reference/api'
apidoc_excluded_paths = ['tests']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'oslo.metrics'
copyright = '2014, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['oslo_metrics.']

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
# html_theme_path = ["."]
html_theme = 'openstackdocs'
# html_static_path = ['static']

# Output file base name for HTML help builder.
htmlhelp_basename = f'{project}doc'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        'index',
        f'{project}.tex',
        f'{project} Documentation',
        'OpenStack Foundation',
        'manual',
    )
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/', None),
    'osloconfig': ('http://docs.openstack.org/oslo.config/latest/', None),
}
